// -*- Mode:c++ -*-
// $Id: MDClient.h,v 1.2 2005/09/25 17:20:44 koblitz Exp $
/** MDClient is a class which provides client functionality for an
    application which wants to run as a client to the metatada
    server. MDClient controls an IP connection to a server providing a
    socket for metadata queries and responses
**/

#ifndef MDCLIENT_H
#define MDCLIENT_H

#include <openssl/ssl.h>
#include <openssl/err.h>

#include <arpa/inet.h>
#include <string>
#include <list>


#define MD_DEFAULT_PORT 8822

class MDClient;
class ConfigParser;

int serverCertVerifyCallback(int ok, X509_STORE_CTX *store);

void int_sig_handler(int sig);
MDClient *registerClient(MDClient * client);

class MDClient {
  friend void int_sig_handler(int);
  friend MDClient *registerClient(MDClient *);
  friend void pthreads_locking_callback(int, int, const char *, int);
 public:
  MDClient(const std::string &host="", int port=0,
	   bool persistent = false,
	   const std::string &configFile = "");

  ~MDClient();

  int connectToServer();

  /// Returns true if the MDClient holds a TCP connection an AMGA server
  bool isConnected() const { return connected; };

  int disconnect(bool saveSession = false);

  int setTimeout(int seconds);

  int execute(const std::string &command);

  int execNoWait(const std::string &command);

  int fetchRow(std::string &row, bool readAhead=true);

  std::string &getError() { return error; };

  bool eot();

  /// Returns the greeting string the server sent on connecting
  const std::string &getGreeting();

  /// Sets the send-buffer when streaming data to the server
  void setSendBuffer(unsigned int s) { sendBufferMax=s; };

  /// Enables debugging of the client to stderr
  void setDebug(bool doit=true) { debug=doit; }

  /// Sets the session number to allow several sessions for a user
  void setSessionNumber(int n) { sessionNumber = n; }

  /// Gets the session number of the user
  int getSessionNumber() { return sessionNumber; }

  int cancel();

  /// Returns the name of the server the client is connected to
  /// (in this session)
  const std::string &serverName() const {
    return sessionHost;
  }

  /// Returns port number of the server the client is connected to
  /// (in this session)
  int serverPort() const {
    return sessionPort;
  }

 private:
  class MDClientLock {
    friend void pthreads_locking_callback(int, int, const char *, int);
  public:
    MDClientLock();
    ~MDClientLock();
  private:
    pthread_mutex_t *locks;
  };

  int prepareExec();
  int setServer(const std::string &host="", int port=0);
  int retrieveResult();

  int tentativeRead();
  int fetchData();
  int doRead();
  void setError(std::string err);
  void sslErrors();
  int sslGetError(int res);

  int establishContext();
  int initConnection(bool &resumeSession);
  int initSSLConnection(bool resumeSession);
  int sendLine(const std::string &c, bool flush=true);
  int initSSL();
  int initSSLSession();
  int readConfiguration(std::list<std::string> &locations);
  int loadCertificate();
  std::string getProxyCertName() const;

  int saveSessionData();
  int restoreSessionData();
  

  static MDClientLock sslLock;

  std::string response;
  int timeout;
  int sock;   // Client socket
  SSL_CTX *ctx;  // SSL server context
  SSL *ssl;
  bool useSSL;
  bool requireSSL;

  std::string error;
  std::string greeting;           // The greeting string of the server
  unsigned int protocolVersion;   // The protocol version of the server

  bool endOfTransmission;
  bool connected;
  ConfigParser *config;
  bool debug;

  int usedPort;

  std::string sendBuffer;
  unsigned int sendBufferMax;
  int resultRows;
  volatile bool cancellationRequested;
  static MDClient *clientRegisteredForInt;
  bool wasCancelled;

  SSL_SESSION *sslSessionData;
  uint64_t sessionID;
  std::string sessionHost;
  int sessionPort;
  int sessionFile;
  std::string sessionFileName;
  bool persistentSessions;
  int sessionNumber;
};

#endif
